@@ -8,6 +8,7 @@ import time
 import re
 import sys
 from datetime import datetime
+from enum import Enum
 from systemd import journal
 
 if sys.version_info.major == 3 and sys.version_info.minor >= 3:
@@ -42,6 +43,14 @@ def get_call_long(call):
     return call_long
 
 
+def get_version_from_lsb():
+    ret, out = run_command("lsb_release -rs")
+    if ret != 0:
+        raise RuntimeError("Cannot get distro version from lsb_release output: '%s'" % out)
+
+    return out.split(".")[0]
+
+
 def get_version_from_pretty_name(pretty_name):
     """ Try to get distro and version from 'OperatingSystemPrettyName'
         hostname property.
@@ -58,7 +67,7 @@ def get_version_from_pretty_name(pretty_name):
     if match is not None:
         version = match.group(0)
     else:
-        raise RuntimeError("Cannot get distro name and version from '%s'" % pretty_name)
+        version = get_version_from_lsb()
 
     return (distro, version)
 
@@ -76,13 +85,12 @@ def get_version():
     if cpe:
         # 2nd to 4th fields from e.g. "cpe:/o:fedoraproject:fedora:25" or "cpe:/o:redhat:enterprise_linux:7.3:GA:server"
         _project, distro, version = tuple(cpe.split(":")[2:5])
+        # we want just the major version, so remove all decimal places (if any)
+        version = str(int(float(version)))
     else:
         pretty_name = str(sys_info.Get("org.freedesktop.hostname1", "OperatingSystemPrettyName", dbus_interface=dbus.PROPERTIES_IFACE))
         distro, version = get_version_from_pretty_name(pretty_name)
 
-    # we want just the major version, so remove all decimal places (if any)
-    version = str(int(float(version)))
-
     return (distro, version)
 
 
@@ -122,12 +130,12 @@ def unstable_test(test):
         except unittest.SkipTest:
             # make sure skipped tests are just skipped as usual
             raise
-        except:
+        except Exception as e:
             # and swallow everything else, just report a failure of an unstable
             # test, unless told otherwise
             if "UNSTABLE_TESTS_FATAL" in os.environ:
                 raise
-            print("unstable-fail...", end="", file=sys.stderr)
+            print("unstable-fail: Ignoring exception '%s'\n" % e, end="", file=sys.stderr)
 
     return decorated_test
 
@@ -179,6 +187,19 @@ class DBusProperty(object):
             else:
                 raise AssertionError('%s != %s' % (self._value, value))
 
+    def assertNotEqual(self, value, timeout=TIMEOUT, getter=None):
+        if getter is not None:
+            check_fn = lambda x: getter(x) != value
+        else:
+            check_fn = lambda x: x != value
+        ret = self._check(timeout, check_fn)
+
+        if not ret:
+            if getter is not None:
+                raise AssertionError('%s == %s' % (getter(self._value), value))
+            else:
+                raise AssertionError('%s == %s' % (self._value, value))
+
     def assertAlmostEqual(self, value, delta, timeout=TIMEOUT, getter=None):
         if getter is not None:
             check_fn = lambda x: abs(getter(x) - value) <= delta
@@ -200,7 +221,7 @@ class DBusProperty(object):
         ret = self._check(timeout, check_fn)
 
         if not ret:
-            raise AssertionError('%s is not grater than %s' % (self._value, value))
+            raise AssertionError('%s is not greater than %s' % (self._value, value))
 
     def assertLess(self, value, timeout=TIMEOUT):
         check_fn = lambda x: x < value
@@ -262,7 +283,12 @@ class DBusProperty(object):
                 raise AssertionError('Expected length %d, but %s has length %d' % (length,
                                                                                    self._value,
                                                                                    len(self._value)))
+    def assertContains(self, member, timeout=TIMEOUT):
+        check_fn = lambda x: member in x
+        ret = self._check(timeout, check_fn)
 
+        if not ret:
+            raise AssertionError('%s does not contain %s' % (self._value, member))
 
 class UdisksTestCase(unittest.TestCase):
     iface_prefix = None
@@ -379,6 +405,18 @@ class UdisksTestCase(unittest.TestCase):
 
         return False
 
+    def try_unmount(self, path):
+        """Handle unmount and retry if busy"""
+        for _ in range(10):
+            ret, out = self.run_command('umount %s' % path)
+            # the mount may either be unmounted already or not exist anymore
+            if ret == 0 or "not mounted" in out or "no mount point specified" in out or "mountpoint not found" in out:
+                return
+            if "target is busy" not in out:
+                break
+            time.sleep(0.5)
+        self.fail('Failed to unmount %s: %s' % (path, out))
+
     @classmethod
     def read_file(self, filename):
         with open(filename, 'r') as f:
@@ -387,11 +425,19 @@ class UdisksTestCase(unittest.TestCase):
 
 
     @classmethod
-    def write_file(self, filename, content, ignore_nonexistent=False):
+    def write_file(self, filename, content, ignore_nonexistent=False, binary=False):
         try:
-            with open(filename, 'w') as f:
+            with open(filename, 'wb' if binary else 'w') as f:
                 f.write(content)
-        except FileNotFoundError as e:
+        except OSError as e:
+            if not ignore_nonexistent:
+                raise e
+
+    @classmethod
+    def remove_file(self, filename, ignore_nonexistent=False):
+        try:
+            os.remove(filename)
+        except OSError as e:
             if not ignore_nonexistent:
                 raise e
 
@@ -399,21 +445,28 @@ class UdisksTestCase(unittest.TestCase):
     def run_command(self, command):
         return run_command(command)
 
+    @classmethod
+    def module_available(cls, module):
+        ret, _out = cls.run_command('modprobe %s' % module)
+        return ret == 0
+
     @classmethod
     def check_module_loaded(self, module):
+        """Tries to load specified module. No checks for extra Manager interface are done.
+           Returns False when module is not available, True when the module initialized
+           successfully, raises an exception otherwise.
+        """
         manager_obj = self.get_object('/Manager')
         manager = self.get_interface(manager_obj, '.Manager')
-        manager_intro = dbus.Interface(manager_obj, "org.freedesktop.DBus.Introspectable")
-        intro_data = manager_intro.Introspect()
-        modules_loaded = 'interface name="org.freedesktop.UDisks2.Manager.%s"' % module in intro_data
-
-        if not modules_loaded:
-            manager.EnableModules(dbus.Boolean(True))
-            intro_data = manager_intro.Introspect()
-            return 'interface name="org.freedesktop.UDisks2.Manager.%s"' % module in intro_data
-        else:
+        try:
+            manager.EnableModule(module, dbus.Boolean(True))
             return True
-
+        except dbus.exceptions.DBusException as e:
+            msg = r"Error initializing module '%s': .*\.so: cannot open shared object file: No such file or directory" % module
+            if re.search(msg, e.get_dbus_message()):
+                return False
+            else:
+                raise
 
     @classmethod
     def ay_to_str(self, ay):
@@ -432,7 +485,22 @@ class UdisksTestCase(unittest.TestCase):
                           signature=dbus.Signature('y'), variant_level=1)
 
     @classmethod
-    def set_udev_property(self, device, prop, value):
+    def bytes_to_ay(self, bytes):
+        """Convert Python bytes to a DBus bytearray"""
+
+        return dbus.Array([dbus.Byte(b) for b in bytes],
+                          signature=dbus.Signature('y'), variant_level=1)
+
+    @classmethod
+    def set_udev_properties(self, device, props):
+        """Sets one or more udev properties for the 'device' identified by its serial number.
+           Note that this overwrites previously set properties. Pass props=None to remove
+           the rules.
+
+        :type props: dict
+        """
+        UDISKS_UDEV_RULES = "/run/udev/rules.d/99-udisks_test.rules"
+
         udev = GUdev.Client()
         dev = udev.query_by_device_file(device)
         serial = dev.get_property("ID_SERIAL")
@@ -443,14 +511,20 @@ class UdisksTestCase(unittest.TestCase):
             # already exists
             pass
 
-        self.write_file("/run/udev/rules.d/99-udisks_test.rules",
-                        'ENV{ID_SERIAL}=="%s", ENV{%s}="%s"\n' % (serial, prop, value))
+        if props:
+            rules = ""
+            for i in props:
+                rules += ', ENV{%s}="%s"' % (i, props[i])
+            self.write_file(UDISKS_UDEV_RULES,
+                            'ENV{ID_SERIAL}=="%s"%s\n' % (serial, rules))
+        else:
+            self.remove_file(UDISKS_UDEV_RULES, ignore_nonexistent=True)
         self.run_command("udevadm control --reload")
         uevent_path = os.path.join(dev.get_sysfs_path(), "uevent")
         self.write_file(uevent_path, "change\n")
         self.udev_settle()
-        os.unlink("/run/udev/rules.d/99-udisks_test.rules")
-        self.run_command("udevadm control --reload")
+        # FIXME: need to give udisksd some time to process the uevent
+        time.sleep(1)
 
     @classmethod
     def assertHasIface(self, obj, iface):
@@ -555,7 +629,7 @@ class JournalRecorder(FlightRecorder):
         self._stopped = monotonic()
 
     def _save(self):
-        j = journal.Reader()
+        j = journal.Reader(converters={"MESSAGE": lambda x: x.decode(errors="replace")})
         j.this_boot()
         j.seek_monotonic(self._started)
         journal_data = ""
@@ -573,3 +647,38 @@ class JournalRecorder(FlightRecorder):
             entry = j.get_next()
         rec = '<<<<< ' + self._desc + ' >>>>>' + '\n' + journal_data + '\n\n\n'
         self._store.append(rec)
+
+
+class TestTags(Enum):
+    ALL = "all"               # "default" tag for running all tests
+    SLOW = "slow"             # slow tests
+    UNSTABLE = "unstable"     # randomly failing tests
+    UNSAFE = "unsafe"         # tests that change system configuration
+    NOSTORAGE = "nostorage"   # tests that don't work with storage
+    EXTRADEPS = "extradeps"   # tests that require special configuration and/or device to run
+
+    @classmethod
+    def get_tags(cls):
+        return [t.value for t in cls.__members__.values()]
+
+    @classmethod
+    def get_tag_by_value(cls, value):
+        tag = next((t for t in cls.__members__.values() if t.value == value), None)
+
+        if not tag:
+            raise ValueError('Unknown value "%s"' % value)
+
+        return tag
+
+
+def tag_test(*tags):
+    def decorator(func):
+        func.slow = TestTags.SLOW in tags
+        func.unstable = TestTags.UNSTABLE in tags
+        func.unsafe = TestTags.UNSAFE in tags
+        func.nostorage = TestTags.NOSTORAGE in tags
+        func.extradeps = TestTags.EXTRADEPS in tags
+
+        return func
+
+    return decorator
