@@ -27,6 +27,8 @@
 #include <string.h>
 #include <stdlib.h>
 #include <sys/sysmacros.h>
+#include <sys/ioctl.h>
+#include <linux/fs.h>
 #include <glib-unix.h>
 
 #include <glib/gstdio.h>
@@ -41,6 +43,7 @@
 #include "udiskslinuxdevice.h"
 #include "udiskslinuxblock.h"
 #include "udiskssimplejob.h"
+#include "udisksstate.h"
 
 /**
  * SECTION:udiskslinuxpartition
@@ -130,8 +133,6 @@ check_authorization (UDisksPartition       *partition,
                                                invocation,
                                                NULL /* GCancellable */,
                                                caller_uid,
-                                               NULL,
-                                               NULL,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
@@ -177,6 +178,38 @@ out:
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+static void update_partitions_list (UDisksObject           *disk_object,
+                                    UDisksLinuxBlockObject *part_object)
+{
+  UDisksPartitionTable *table = NULL;
+  gchar **partitions = NULL;
+  const gchar *object_path = NULL;
+  guint num_parts = 0;
+
+  object_path = g_dbus_object_get_object_path (G_DBUS_OBJECT (part_object));
+  if (object_path == NULL)
+    return;
+
+  table = udisks_object_peek_partition_table (disk_object);
+  if (table == NULL)
+    return;
+
+  partitions = udisks_partition_table_dup_partitions (table);
+  if (partitions != NULL && g_strv_contains ((const gchar **) partitions, object_path))
+    /* this partition is already in the property */
+    goto out;
+
+  num_parts = partitions ? g_strv_length (partitions) : 0;
+  partitions = g_realloc (partitions, (num_parts + 2) * sizeof (gchar *));
+  partitions[num_parts] = g_strdup (object_path);
+  partitions[num_parts + 1] = NULL;
+
+  udisks_partition_table_set_partitions (table, (const gchar**) partitions);
+
+out:
+  g_strfreev (partitions);
+}
+
 /**
  * udisks_linux_partition_update:
  * @partition: A #UDisksLinuxPartition.
@@ -210,7 +243,7 @@ udisks_linux_partition_update (UDisksLinuxPartition   *partition,
       type = g_udev_device_get_property (device->udev_device, "ID_PART_ENTRY_TYPE");
       offset = g_udev_device_get_property_as_uint64 (device->udev_device, "ID_PART_ENTRY_OFFSET") * G_GUINT64_CONSTANT (512);
       size = g_udev_device_get_property_as_uint64 (device->udev_device, "ID_PART_ENTRY_SIZE") * G_GUINT64_CONSTANT (512);
-      name = udisks_decode_udev_string (g_udev_device_get_property (device->udev_device, "ID_PART_ENTRY_NAME"));
+      name = udisks_decode_udev_string (g_udev_device_get_property (device->udev_device, "ID_PART_ENTRY_NAME"), NULL);
       uuid = g_udev_device_get_property (device->udev_device, "ID_PART_ENTRY_UUID");
       flags = g_udev_device_get_property_as_uint64 (device->udev_device, "ID_PART_ENTRY_FLAGS");
 
@@ -263,8 +296,10 @@ udisks_linux_partition_update (UDisksLinuxPartition   *partition,
         }
     }
 
-  if (disk_block_object != NULL)
+  if (disk_block_object != NULL) {
     table_object_path = g_dbus_object_get_object_path (G_DBUS_OBJECT (disk_block_object));
+    update_partitions_list (disk_block_object, object);
+  }
 
   udisks_partition_set_number (UDISKS_PARTITION (partition), number);
   udisks_partition_set_type_ (UDISKS_PARTITION (partition), type);
@@ -277,6 +312,8 @@ udisks_linux_partition_update (UDisksLinuxPartition   *partition,
   udisks_partition_set_is_container (UDISKS_PARTITION (partition), is_container);
   udisks_partition_set_is_contained (UDISKS_PARTITION (partition), is_contained);
 
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (partition));
+
   g_free (name);
   g_clear_object (&device);
   g_clear_object (&disk_block_object);
@@ -293,6 +330,7 @@ handle_set_flags (UDisksPartition       *partition,
   UDisksBlock *block = NULL;
   UDisksObject *object = NULL;
   UDisksDaemon *daemon = NULL;
+  UDisksState *state = NULL;
   gchar *device_name = NULL;
   UDisksObject *partition_table_object = NULL;
   UDisksPartitionTable *partition_table = NULL;
@@ -318,6 +356,7 @@ handle_set_flags (UDisksPartition       *partition,
     }
 
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
+  state = udisks_daemon_get_state (daemon);
   block = udisks_object_get_block (object);
   partition_table_object = udisks_daemon_find_object (daemon, udisks_partition_get_table (partition));
   partition_table = udisks_object_get_partition_table (partition_table_object);
@@ -325,6 +364,9 @@ handle_set_flags (UDisksPartition       *partition,
   device_name = udisks_block_dup_device (partition_table_block);
   partition_name = udisks_block_dup_device (block);
 
+  udisks_linux_block_object_lock_for_cleanup (UDISKS_LINUX_BLOCK_OBJECT (object));
+  udisks_state_check_block (state, udisks_linux_block_object_get_device_number (UDISKS_LINUX_BLOCK_OBJECT (object)));
+
   /* hold a file descriptor open to suppress BLKRRPART generated by the tools */
   fd = open (partition_name, O_RDONLY);
 
@@ -345,6 +387,8 @@ handle_set_flags (UDisksPartition       *partition,
     {
       if (flags & 1) /* 1 << 0 */
           bd_flags |= BD_PART_FLAG_GPT_SYSTEM_PART;
+      if (flags & 4) /* 1 << 2 */
+          bd_flags |= BD_PART_FLAG_LEGACY_BOOT;
       if (flags & 0x1000000000000000) /* 1 << 60 */
           bd_flags |= BD_PART_FLAG_GPT_READ_ONLY;
       if (flags & 0x4000000000000000) /* 1 << 62 */
@@ -399,13 +443,18 @@ handle_set_flags (UDisksPartition       *partition,
         goto out;
       }
 
-  udisks_linux_block_object_trigger_uevent (UDISKS_LINUX_BLOCK_OBJECT (object));
+  udisks_linux_block_object_trigger_uevent_sync (UDISKS_LINUX_BLOCK_OBJECT (object),
+                                                 UDISKS_DEFAULT_WAIT_TIMEOUT);
   udisks_partition_complete_set_flags (partition, invocation);
   udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), TRUE, NULL);
 
  out:
   if (fd != -1)
     close (fd);
+  if (object != NULL)
+    udisks_linux_block_object_release_cleanup_lock (UDISKS_LINUX_BLOCK_OBJECT (object));
+  if (state != NULL)
+    udisks_state_check (state);
   g_free (device_name);
   g_free (partition_name);
   g_clear_error (&error);
@@ -431,6 +480,7 @@ handle_set_name (UDisksPartition       *partition,
   UDisksBlock *block = NULL;
   UDisksObject *object = NULL;
   UDisksDaemon *daemon = NULL;
+  UDisksState *state = NULL;
   gchar *device_name = NULL;
   gchar *partition_name = NULL;
   UDisksObject *partition_table_object = NULL;
@@ -454,8 +504,12 @@ handle_set_name (UDisksPartition       *partition,
     }
 
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
+  state = udisks_daemon_get_state (daemon);
   block = udisks_object_get_block (object);
 
+  udisks_linux_block_object_lock_for_cleanup (UDISKS_LINUX_BLOCK_OBJECT (object));
+  udisks_state_check_block (state, udisks_linux_block_object_get_device_number (UDISKS_LINUX_BLOCK_OBJECT (object)));
+
   partition_table_object = udisks_daemon_find_object (daemon, udisks_partition_get_table (partition));
   partition_table = udisks_object_get_partition_table (partition_table_object);
   partition_table_block = udisks_object_get_block (partition_table_object);
@@ -516,13 +570,18 @@ handle_set_name (UDisksPartition       *partition,
       goto out;
     }
 
-  udisks_linux_block_object_trigger_uevent (UDISKS_LINUX_BLOCK_OBJECT (object));
+  udisks_linux_block_object_trigger_uevent_sync (UDISKS_LINUX_BLOCK_OBJECT (object),
+                                                 UDISKS_DEFAULT_WAIT_TIMEOUT);
   udisks_partition_complete_set_name (partition, invocation);
   udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), TRUE, NULL);
 
  out:
   if (fd != -1)
     close (fd);
+  if (object != NULL)
+    udisks_linux_block_object_release_cleanup_lock (UDISKS_LINUX_BLOCK_OBJECT (object));
+  if (state != NULL)
+    udisks_state_check (state);
   g_free (device_name);
   g_free (partition_name);
   g_clear_error (&error);
@@ -752,6 +811,36 @@ handle_set_type (UDisksPartition       *partition,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+
+typedef struct
+{
+  const gchar *object_path;
+  guint64      new_size;
+} WaitForPartitionResizeData;
+
+static UDisksObject *
+wait_for_partition_resize (UDisksDaemon *daemon,
+                           gpointer      user_data)
+{
+  WaitForPartitionResizeData *data = user_data;
+  UDisksObject *object = NULL;
+  UDisksPartition *partition = NULL;
+
+  object = udisks_daemon_find_object (daemon, data->object_path);
+
+  if (object != NULL)
+    {
+      partition = udisks_object_peek_partition (object);
+      if (udisks_object_peek_block (object) == NULL ||
+          partition == NULL || udisks_partition_get_size (partition) != data->new_size)
+        {
+          g_clear_object (&object);
+        }
+    }
+
+  return object;
+}
+
 /* runs in thread dedicated to handling @invocation */
 static gboolean
 handle_resize (UDisksPartition       *partition,
@@ -762,11 +851,16 @@ handle_resize (UDisksPartition       *partition,
   UDisksBlock *block = NULL;
   UDisksObject *object = NULL;
   UDisksDaemon *daemon = NULL;
+  UDisksState *state = NULL;
   UDisksObject *partition_table_object = NULL;
   UDisksBlock *partition_table_block = NULL;
   uid_t caller_uid;
   GError *error = NULL;
   UDisksBaseJob *job = NULL;
+  UDisksObject *partition_object = NULL;
+  WaitForPartitionResizeData wait_data;
+  gint fd = 0;
+  const gchar *part = NULL;
 
   if (!check_authorization (partition, invocation, options, &caller_uid))
     {
@@ -780,11 +874,18 @@ handle_resize (UDisksPartition       *partition,
       goto out;
     }
 
+  wait_data.object_path = g_dbus_object_get_object_path (G_DBUS_OBJECT (object));
+  wait_data.new_size = 0; /* hit timeout in case of error */
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
+  state = udisks_daemon_get_state (daemon);
   block = udisks_object_get_block (object);
+  part = udisks_block_get_device (block);
   partition_table_object = udisks_daemon_find_object (daemon, udisks_partition_get_table (partition));
   partition_table_block = udisks_object_get_block (partition_table_object);
 
+  udisks_linux_block_object_lock_for_cleanup (UDISKS_LINUX_BLOCK_OBJECT (object));
+  udisks_state_check_block (state, udisks_linux_block_object_get_device_number (UDISKS_LINUX_BLOCK_OBJECT (object)));
+
   job = udisks_daemon_launch_simple_job (daemon,
                                          UDISKS_OBJECT (object),
                                          "partition-modify",
@@ -812,13 +913,46 @@ handle_resize (UDisksPartition       *partition,
       goto out;
     }
 
+  /* Wait for partition property to be updated so that the partition interface
+   * will not disappear shortly after this method returns.
+   * Clients could either explicitly wait for an interface or try
+   * udisks_client_settle() to wait for interfaces to be present.
+   * If the partition size wasn't changed then there won't be any reappearing
+   * of the partition node or the interfaces.
+   */
+
+  fd = open (part, O_RDONLY);
+  if (fd != -1) {
+    if (ioctl (fd, BLKGETSIZE64, &wait_data.new_size) == -1) {
+        udisks_warning ("Could not query new partition size for %s", part);
+    }
+
+    close (fd);
+  } else {
+    udisks_warning ("Could not open %s to query new partition size", part);
+  }
+
+  udisks_linux_block_object_trigger_uevent_sync (UDISKS_LINUX_BLOCK_OBJECT (partition_table_object ? partition_table_object : object),
+                                                 UDISKS_DEFAULT_WAIT_TIMEOUT);
+  partition_object = udisks_daemon_wait_for_object_sync (daemon,
+                                                         wait_for_partition_resize,
+                                                         &wait_data,
+                                                         NULL,
+                                                         UDISKS_DEFAULT_WAIT_TIMEOUT,
+                                                         NULL);
+
   udisks_partition_complete_resize (partition, invocation);
   udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), TRUE, NULL);
 
  out:
+  if (object != NULL)
+    udisks_linux_block_object_release_cleanup_lock (UDISKS_LINUX_BLOCK_OBJECT (object));
+  if (state != NULL)
+    udisks_state_check (state);
   g_clear_error (&error);
   g_clear_object (&object);
   g_clear_object (&block);
+  g_clear_object (&partition_object);
   g_clear_object (&partition_table_object);
   g_clear_object (&partition_table_block);
 
@@ -837,6 +971,7 @@ handle_delete (UDisksPartition       *partition,
   UDisksBlock *block = NULL;
   UDisksObject *object = NULL;
   UDisksDaemon *daemon = NULL;
+  UDisksState *state = NULL;
   gchar *device_name = NULL;
   gchar *partition_name = NULL;
   UDisksObject *partition_table_object = NULL;
@@ -861,10 +996,14 @@ handle_delete (UDisksPartition       *partition,
     }
 
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
+  state = udisks_daemon_get_state (daemon);
   block = udisks_object_get_block (object);
   partition_table_object = udisks_daemon_find_object (daemon, udisks_partition_get_table (partition));
   partition_table_block = udisks_object_get_block (partition_table_object);
 
+  udisks_linux_block_object_lock_for_cleanup (UDISKS_LINUX_BLOCK_OBJECT (object));
+  udisks_state_check_block (state, udisks_linux_block_object_get_device_number (UDISKS_LINUX_BLOCK_OBJECT (object)));
+
   if (teardown_flag)
     {
       if (!udisks_linux_block_teardown (block, invocation, options, &error))
@@ -905,12 +1044,17 @@ handle_delete (UDisksPartition       *partition,
       goto out;
     }
   /* this is sometimes needed because parted(8) does not generate the uevent itself */
-  udisks_linux_block_object_trigger_uevent (UDISKS_LINUX_BLOCK_OBJECT (partition_table_object));
+  udisks_linux_block_object_trigger_uevent_sync (UDISKS_LINUX_BLOCK_OBJECT (partition_table_object),
+                                                 UDISKS_DEFAULT_WAIT_TIMEOUT);
 
   udisks_partition_complete_delete (partition, invocation);
   udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), TRUE, NULL);
 
  out:
+  if (object != NULL)
+    udisks_linux_block_object_release_cleanup_lock (UDISKS_LINUX_BLOCK_OBJECT (object));
+  if (state != NULL)
+    udisks_state_check (state);
   g_free (device_name);
   g_free (partition_name);
   g_clear_error (&error);
